"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/*
 * SonarTS
 * Copyright (C) 2017-2017 SonarSource SA
 * mailto:info AT sonarsource DOT com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
const tslint = require("tslint");
const ts = require("typescript");
const cfg_1 = require("../cfg/cfg");
const nav = require("../utils/navigation");
class Rule extends tslint.Rules.AbstractRule {
    apply(sourceFile) {
        return this.applyWithWalker(new Walker(sourceFile, this.getOptions()));
    }
}
Rule.metadata = {
    description: "Jump statements should not be used unconditionally",
    options: null,
    optionsDescription: "",
    rspecKey: "RSPEC-1751",
    ruleName: "no-unconditional-jump",
    type: "functionality",
    typescriptOnly: false,
};
exports.Rule = Rule;
class Walker extends tslint.RuleWalker {
    visitBreakStatement(node) {
        this.checkJump(node);
    }
    visitContinueStatement(node) {
        this.checkJump(node);
    }
    visitThrowStatement(node) {
        this.checkJump(node);
    }
    visitReturnStatement(node) {
        this.checkJump(node);
    }
    checkJump(node) {
        if (this.isConditional(node))
            return;
        if (node.kind === ts.SyntaxKind.BreakStatement && this.isInsideForIn(node))
            return;
        if (node.kind === ts.SyntaxKind.ReturnStatement && this.isInsideForOf(node))
            return;
        const cfg = this.buildCfg(node);
        if (!cfg)
            return;
        const keyword = nav.keyword(node);
        const loop = nav.firstLocalAncestor(node, ...nav.LOOP_STATEMENTS);
        if (!loop)
            return;
        if (node.kind === ts.SyntaxKind.ContinueStatement) {
            this.raiseIssue(keyword);
        }
        else {
            const loopingBlock = cfg.findLoopingBlock(loop);
            if (loopingBlock && this.actuallyLoops(loopingBlock))
                return;
            this.raiseIssue(keyword);
        }
    }
    raiseIssue(keyword) {
        this.addFailureAtNode(keyword, `Remove this "${keyword.getText()}" statement or make it conditional`);
    }
    isConditional(node) {
        const parents = nav.localAncestorsChain(node).map(p => p.kind);
        const conditionalsAndLoops = parents.filter(kind => nav.LOOP_STATEMENTS.concat(nav.CONDITIONAL_STATEMENTS).includes(kind));
        return conditionalsAndLoops.length > 0 && nav.CONDITIONAL_STATEMENTS.includes(conditionalsAndLoops[0]);
    }
    isInsideForIn(node) {
        const parentLoop = nav.firstLocalAncestor(node, ...nav.LOOP_STATEMENTS);
        return !!parentLoop && parentLoop.kind === ts.SyntaxKind.ForInStatement;
    }
    isInsideForOf(node) {
        const parentLoop = nav.firstLocalAncestor(node, ...nav.LOOP_STATEMENTS);
        return !!parentLoop && parentLoop.kind === ts.SyntaxKind.ForOfStatement;
    }
    buildCfg(node) {
        const wrappingFunction = nav.firstLocalAncestor(node, ...nav.FUNCTION_LIKE);
        if (wrappingFunction && wrappingFunction.body) {
            if (wrappingFunction.body.kind === ts.SyntaxKind.Block) {
                return cfg_1.ControlFlowGraph.fromStatements(wrappingFunction.body.statements);
            }
            else {
                return; // When moving buildCfg to cfg file this should be replaced by fromExpression, here instead, we skip
            }
        }
        return cfg_1.ControlFlowGraph.fromStatements(node.getSourceFile().statements);
    }
    actuallyLoops(block) {
        if (!block.loopingStatement)
            return false;
        const loopContents = this.collectLoopContents(block.loopingStatement);
        if (block instanceof cfg_1.CfgBlockWithPredecessors) {
            return !!block.predecessors.find(predecessor => this.hasPredecessor(predecessor) &&
                !!predecessor.getElements().find(predecessorElement => loopContents.includes(predecessorElement)));
        }
        else {
            return false;
        }
    }
    collectLoopContents(iterationStatement) {
        const bodyContents = nav.descendants(iterationStatement.statement);
        if (iterationStatement.kind === ts.SyntaxKind.ForStatement) {
            const updateExpression = iterationStatement.incrementor;
            if (updateExpression)
                return bodyContents.concat(nav.descendants(updateExpression));
        }
        return bodyContents;
    }
    hasPredecessor(block) {
        if (block instanceof cfg_1.CfgBlockWithPredecessors) {
            return block.predecessors.length > 0;
        }
        else {
            return false;
        }
    }
}
//# sourceMappingURL=noUnconditionalJumpRule.js.map